#include "colon-variable.h"
#include "throf.h"
#include "memory.h"
#include "function.h"
#include "stack.h"
#include "colon-state.h"
#include "error.h"
#include "colon-constants.h"

/* for now at least it is illegal for a variable to 
   be in multiple registers */


/*------------------- fundamental reg_state operations ----------------*/
void init_state(reg_state *state)
{
  uval i;
  state->var_list = NIL;
  for(i = 0; i < MAX_REG; i++)
    state->regs[i] = NIL;
}

/*-------------- tests for basic register operations -------------- */
int remove_ok(varbl *var)
{
  int value;
  static uval test1 = flushed | safe_to_remove;
  static uval test2 = safe_to_flush | safe_to_remove;
  value = (var->flags & test1) == test1;
  value = value || ((var->flags & test2) == test2);
  return value;
}

int move_ok(varbl *var)
{
  return (var->reg == ZERO) || (var->flags & safe_to_move);
}

/*------------------- basic regiser operations -----------------------*/
/* I've got so many flags I've to factor this */

void normalize_flags(varbl *var);
void flush_reg(varbl *var);
void remove_reg(reg_state *state, varbl *var);
void load_reg(reg_state *state,varbl *var,reg_t reg);
void move_reg(reg_state *state,varbl *var,reg_t reg);

/* if the flags are set somewhat inconsitently allow
   the most permissive diffinition */
void normalize_flags(varbl *var)
{
  /* this is the only weird case I can conceive of */
  if ((var->flags & (safe_to_flush | safe_to_remove)) || 
    (var->flags & (flushed  | safe_to_remove)))
    {
      var->flags  |= safe_to_move;
    }
}

void flush_reg(varbl *var)
{
  normalize_flags(var);
  if (var->reg != ZERO) /* if it's zero it's already flushed */
    {
      if (!(var->flags & flushed))  /* if it's flushed it needn't be flushed */
	{
	  if (var->flags & safe_to_flush)
	    {
	      var->flags &= ~safe_to_flush; /* prevent infinite recursion */
	      ppush(var);
	      var->save();
	      var->flags |= flushed | safe_to_flush;
	    }
	  else
	    {
	      char string[] = "attempt to flush an unflushable register, aborting!!\n";
	      dpush(sizeof(string) -1);
	      ppush(string);
	      VVfatal_error();
	    }
	}
    }
}

void remove_reg(reg_state *state, varbl *var)
{
  normalize_flags(var);
  if (var->reg == ZERO)
    return;
  else if (var->flags & safe_to_remove)
    {
      flush_reg(var);
      state->regs[var->reg] = NIL;
      var->reg = ZERO;
      var->flags = safe_to_flush | flushed | safe_to_move | safe_to_remove;
    }
  else
    {
      char string[] = "attempt to remove an unremoveable register, aborting!!\n";
      dpush(sizeof(string) -1);
      ppush(string);
      VVfatal_error();
    }
}

void load_reg(reg_state *state, varbl *var, reg_t reg)
{
  normalize_flags(var);
  if (var->reg != ZERO) {
    move_reg(state, var, reg);   /* if it's already in a register move it */
  }
  else if (reg != ZERO)  /* loading to ZERO is a noop */
    {
      /* here if the destination is busy, I should attempt to move the */
      /* variable if I can't remove it */
      if (state->regs[reg] != NIL)
	remove_reg(state,state->regs[reg]);
      var->reg = reg;
      state->regs[reg] = var;
      var->flags = safe_to_flush | flushed | safe_to_move | safe_to_remove;
      var->flags &= ~safe_to_remove;
      ppush(var);
      var->load();
      var->flags |= safe_to_remove;
    }
}

void move_reg(reg_state *state, varbl *var, reg_t reg)
{
  normalize_flags(var);
  if (var->reg == ZERO)
    {
      load_reg(state, var, reg); /* if the variable is not in a register load it */
    }
  else if (reg == ZERO)
    remove_reg(state,var);  /* moving to zero is removing the register */
  else if (var->reg == reg)
    { ; }
  else if (var->flags & safe_to_move)
    {
      reg_t reg2;
      varbl *var2;
      op_t *op;

      var2 = state->regs[reg];
      reg2 = var->reg;
      normalize_flags(var2);

      if (!(var2->flags & safe_to_move))  /* if I can't move it remove it */
	{
	  remove_reg(state, var2);
	  var2 = NIL;  /* pretend it wasn't here in the first place */
	}

      op = allocate(sizeof(op_t));
      op->opcode = XCHG;
      op->r1 = var->reg;
      op->r2 = reg;
      ppush(op);  VVappend_def();
      
      var->reg = reg;
      state->regs[reg] = var;
      
      if (var2 != NIL) var2->reg = reg2;
      state->regs[reg2] = var2;
    }
  else
    {
      char string[] = "attempt to move an unmoveable register, aborting!!\n";
      dpush(sizeof(string) -1);
      ppush(string);
      VVfatal_error();
    }
}

/*------------- register_list routines -----------------*/

void add_to_list(reg_state *state, var_node *new)
{
  var_node *start;
  start = state->var_list;
  new->next = NIL;
  new->prev = NIL;
  if (start == NIL)
    {
      state->var_list = new;
    }
  else
    {
      while(start->next != NIL)
	start = start->next;
      start->next = new;
      new->prev = start;
    }
}


var_node * find_match(reg_state *state,var_node *tmpl)
{
  var_node *iter;
  iter = state->var_list;
  if (iter == NIL)
    return NIL;
  do {
    if ((iter->var.u == tmpl->var.u) && (iter->var.v == tmpl->var.v) && 
	(iter->var.load == tmpl->var.load) && (iter->var.save == tmpl->var.save) &&
	(iter->var.preserve == tmpl->var.preserve))
      {
	return iter;
      }
    iter = iter->next;
  }while(iter != NIL);
  return NIL;
}

/*-------------- allocated register routines -------------*/

varbl *get_safe_to_remove(reg_state *state)
{
  varbl * ret;
  uval i;
  for(i = 0; i < MAX_REG; i++)
    {
      if (remove_ok(state->regs[i]))
	{
	  ret = state->regs[i];
	  break;
	}
    }
  if (i == MAX_REG)
    ret = 0;
  return ret;
}

reg_t get_free_reg(reg_state *state)
{
  uval i;
  for(i = 0; i < MAX_REG; i++)
    {
      if (state->regs[i] == NIL)
	return i;
    }
  return ZERO;
}

reg_t aquire_free_reg(reg_state *state)  /* get a free register or make one */
{
  varbl *safe;
  reg_t ret;
  ret = get_free_reg(state);
  if (ret == ZERO)
    {
      safe = get_safe_to_remove(state);
      if (safe == NIL)
	{
	  char string[] = "attempt to allocate a register and none avaiable, aborting!!\n";
	  dpush(sizeof(string) -1);
	  ppush(string);
	  VVfatal_error();
	}
      else
	{
	  ret = safe->reg;
	  remove_reg(state,safe);  
	}
    }
  return ret;
}

/*------------------- interface routines -------------------*/
varbl * new_variable(reg_state *state, variable_method load, 
		     variable_method save, variable_method preserve, uval u, void *v)
{
  var_node template;
  var_node * new;

/* template is used to find and identical node if present in already in the list */
  template.next = NIL;
  template.prev = NIL;
  template.var.u = u;
  template.var.v = v;
  template.var.reg = ZERO;
  template.var.load = load;
  template.var.save = save;
  template.var.preserve = preserve;
  template.var.flags = safe_to_flush | flushed | safe_to_move | safe_to_remove;

  new = find_match(state,&template);
  if (new == NIL)
    {
      new = allocate(sizeof(var_node));
      *new = template; /* structure copy */
      add_to_list(state,new);
    }
  return &new->var;
}

reg_t reg_for_variable(reg_state *state,varbl *var)
{
  if (var->reg == ZERO)
    {
      reg_t reg;
      reg = aquire_free_reg(state);
      load_reg(state, var, reg);
    }
  var->flags &= ~(safe_to_remove | flushed);
  return var->reg;
}

void assign_variable(reg_state *state, varbl *var, reg_t reg)
{
  move_reg(state, var, reg);
  var->flags &= ~(safe_to_remove | safe_to_move | flushed);
}

/* states that magically a variable has gotten into an empty register */
void in_variable(reg_state *state, varbl *var, reg_t reg)
{
  var->flags &= ~(safe_to_remove | flushed);
  if (state->regs[reg] == var) return; /* all work has already been done */
  if (state->regs[reg] != NIL)
    {
      char string[] = "two variables in the same register, aborting!!\n";
      dpush(sizeof(string) -1);
      ppush(string);
      VVfatal_error();
    }
  else
    {
      if (var->reg != ZERO)
	state->regs[var->reg] = NIL;
      state->regs[reg] = var;
      var->reg = reg;
    }
}

/* transfer values from one variable to another */
void transfer_value(reg_state *state, varbl *from, varbl *to)
{
  if (from->reg == ZERO)  /* executing here implies remove_ok */
    {
      reg_t reg;
      reg = aquire_free_reg(state);
      load_reg(state, from, reg); 
    }
  if (remove_ok(from) && move_ok(to))
      {
	flush_reg(from);
	state->regs[from->reg] = to;
	if(to->reg != ZERO)
	  state->regs[to->reg] = NIL;
	to->reg = from->reg;
	from->reg = ZERO;
	to->flags &= ~flushed;
      }
  else
    {
      op_t *op;
      if (to->reg == ZERO)
	{
	  reg_t reg;
	  reg = aquire_free_reg(state);  /* this won't kick from out */
	  state->regs[reg] = to;         /* cause from can't be removed */
	  to->reg = reg;
	}
      op = allocate(sizeof(op_t));
      op->opcode = SET;
      op->r1 = to->reg;
      op->r2 = from->reg;
      op->value = 0;
      ppush(op);
      VVappend_def();

      to->flags &= ~flushed;
    }
}

void release_variable(reg_state *state, varbl *var)
{
  /* I'm not sure this is the correct combinations of flags */
  /* I'm really not sure about safe_to_flush */
  /* I won't include safe_to_flush for now */
  var->flags |= safe_to_move | safe_to_remove ;
}

void flush_variable(reg_state *state, varbl *var)
{
  flush_reg(var);
}

/* say that any contents of the register stored in a register are invalid */
void invalidate_variable(reg_state *state, varbl *var)
{
  if (var->reg != ZERO)
    {
      state->regs[var->reg] = NIL;
      var->reg = ZERO;
    }
}

/* say that any contents of the register stored in a register are invalid */
void invalidate_all(reg_state *state)
{
  uval i;
  for(i = 0; i < MAX_REG; i++)
    {
      if (state->regs[i] != NIL)
	{
	  state->regs[i]->reg = ZERO;
	  state->regs[i] = NIL;
	}
    }
}


arg_array *get_register_state(reg_state *state)
{
  uval i;
  arg_array *array = allocate(sizeof(arg_array));
/*  static arg_t random = { RANDOM_REG, {{ 0, 0 }}}; */

  for(i = 0; i < MAX_REG; i++)
    {
      if (state->regs[i] == NIL)   /* make sure there is a variable assigned */
	{                          /* so nothing else uses that register */
	  varbl *var;
	  dpush(i); VVget_dummy(); var = ppop();
	  assign_variable(state, var, i); 
	}
      ppush(&array->a[i]); ppush(state->regs[i]); state->regs[i]->preserve();
    }
  for(i = 0; i < MAX_REG; i++)
    {
      if(array->a[i].type == RANDOM_REG)
	invalidate_variable(state,state->regs[i]);
    }

  return array;
}

