
/* This file implements a very basic colon compiler for a forth-like. */

/* It implements essentially procedure threading, except that it has a */
/* goto construct.  It's virtual machine has the following operations: */

#include <stdio.h>
#include <stdlib.h>
#include "throf.h"
#include "compile.h"
#include "memory.h"
#include "function.h"

#define EXPLICIT_NOPS 1

op_t *compile_operation(op_t * op);
void fixup_operation(op_t *op);

#define CHUNK_SIZE 32

typedef union vals {
  void * address;
  unsigned value;
  char data[4];
}addr_un;


uchar * first = NIL;
uchar * place = NIL;
uchar * last = NIL;

uval pass;

uval offset(void)
{
  return place - first;
}

uval free_space(void)
{
  return last - place;
}


static void init_write_mem(void)
{
  if (first == NIL)
    {
      first = allocate(CHUNK_SIZE);
      place = first;
      last = first + CHUNK_SIZE-1;
    }
}

static void kill_write_mem(void)
{
  grow (-free_space(),first);
  first = NIL;
  place = NIL;
  last = NIL;
}
/* compile_word 
   is not reentrant!
 */

/* for the moment no functions
   use the optimization of passing variables
   or the stack in registers.
   This will come eventually.
*/
func_ptr_t compile_word(word_t * word)
{
  op_t * op;    
  uval offset;
  func_ptr_t func;

  init_write_mem();  /* so I have a place to write the function */

  /* first pass set all offsets to a know value */
  pass = 1;
  op = word->first_op;
  while(op != NIL)             /* I can probably take this pass out */
    {
      op->offset = UVAL_MAX;
      op = op->next;
    }

  /* second pass ouput the instructions */
  /* third pass do the fix_ups */
  for(pass = 2; pass < 4; pass++)
    {
      place = first;
      offset = 0;
      op = word->first_op;
      while(op != NIL)
	op = compile_operation(op);  /* returns the next instruction */
    }

  func = (func_ptr_t)(first);
  kill_write_mem();  /* I'm done using it, remember that */
  
  return func;
}

void write_byte(uchar byte)
{
  *place = byte;
  if (place == last)
    {
      grow(CHUNK_SIZE, first);
      last += CHUNK_SIZE;
    }
  place++;
}

typedef void (*wb_t)(uchar);
typedef void (*comp_func_t)(wb_t, op_t *);
typedef op_t *(*test_func_t)(op_t *);

void compile_ret(wb_t, op_t *);
void compile_call(wb_t, op_t *);
void compile_jmp(wb_t, op_t *);
void compile_gotoq(wb_t, op_t *);
void compile_load(wb_t, op_t *);
void compile_store(wb_t, op_t *);
void compile_set(wb_t, op_t *);
void compile_xchg(wb_t, op_t *);
void compile_add(wb_t, op_t *);
void compile_sub(wb_t, op_t *);
void compile_syscall(wb_t, op_t *);
void compile_nop(wb_t, op_t *);
void output_nop(wb_t, op_t *);

#define EXCEPTIONS 2
op_t *is_pushl(op_t *op);
void compile_pushl(wb_t, op_t *);
op_t *is_popl(op_t *op);
void compile_popl(wb_t, op_t *);

op_t *compile_operation(op_t * op)  /* returns the next opcode :) */
{     /* order is very important here!! */
  static comp_func_t compile_funcs[LAST_OPCODE] = {
    compile_ret,      compile_call,     compile_jmp,      compile_gotoq,
    compile_load,     compile_store,    compile_set,      compile_xchg,     
    compile_add,      compile_sub,      compile_syscall,  compile_nop
  };
  static test_func_t test_funcs[EXCEPTIONS] = {
    is_pushl,         is_popl
  };
  static comp_func_t exception_funcs[EXCEPTIONS] = {
    compile_pushl,     compile_popl
  };
  op_t * next;

  op->offset = offset();
  next = op->next;
  if (op->opcode < LAST_OPCODE)
    {
      comp_func_t func;
      int i;
      func = compile_funcs[op->opcode];
      for(i = 0; i < EXCEPTIONS; i++) { 
	  op_t *maybe_next;
	  maybe_next = test_funcs[i](op);
	  if(maybe_next) {
	    next = maybe_next;
	    func = exception_funcs[i];
	    break;
	  }
	}
      func(write_byte,op);
    }
  else
    {
      fprintf(stderr,"Invalid opcode passed to the compiler, aborting!!\n");
      exit(EXIT_FAILURE);
    }
  /* in the second pass correct for extra long size estimates */
  if ((op->next != NIL) && (op->next->offset != UVAL_MAX))
    {
      while(offset() != op->next->offset)
	{
	  output_nop(write_byte,op);
	}
    }
  return next;
}

typedef enum { contents_of = 0, contents_disp8_reg = 1, contents_disp32_reg =2 ,
	 value_of = 3} mod_t;

typedef enum { times1 = 0, times2 = 1, times4 = 2, times8 = 3} scale_t;

enum arop { add, or, adc, sbb, and, sub, xor, cmp };

#define disp32 5
#define scaled_index 4
#define no_base 5
#define no_index 4

/* note I have switched the order of reg & r/m from the
   `natural' machine order of mod reg r/m because mod & r/m
    are stronly associated and reg is basically ignored */

unsigned char calc_mod_byte(mod_t mod, reg_t r_m, reg_t reg)
{
  r_m &= 7;
  reg &= 7;
  mod &= 3;
  return (mod << 6) | (reg << 3) | (r_m);
}

unsigned char calc_scale_byte(scale_t scale, reg_t index, reg_t base)
{
  return (scale << 6) | ( index << 3) | (base );
}

inline int is_short_offset(uval val)   {  return (val <= 127 ) || (val >= 0xFFFFFF80); }
inline uchar short_offset(uval val)    {  return (val&0xFF); }
inline int is_medium_offset(uval val) 
{  
  return !is_short_offset && ((val <= 0x7FFF) || (val >= 0xFFFF8001));
}
inline uval to_medium_offset(uval val) {  return val & 0xFFFF; }
inline int is_long_offset(uval val)    {  return !is_short_offset(val); }
inline uval long_offset(uval val)      {  return val; }



/* length 5 */
void compile_call(wb_t wb, op_t *op)
{                                    /* I have a problem the call is relative */
  addr_un repack;                    /* I don't know how I'm going to GC it */

  repack.value = ((char *)op->word->func) - ((char *)place +5);  /* relocate address */
                                      /* using place ^^^^^ here is a hack */
  wb(0xE8);               
  wb(repack.data[0]);  
  wb(repack.data[1]);   
  wb(repack.data[2]);
  wb(repack.data[3]);
}

/* offset are measured from the beginning of this instruction! */
/* that is before the instruction is written */
void compile_jmp(wb_t wb, op_t  * op)
{
  addr_un repack;  /* I won't use the short for of jump,
                      because of reloacation of words */

  repack.value = ((char *)op->word->func) - ((char *)place +5);
                                      /* using place ^^^^^^ here is a hack */

  wb(0xe9);
  wb(repack.data[0]);
  wb(repack.data[1]);
  wb(repack.data[2]);
  wb(repack.data[3]);
  
}

/* offset are measured from the beginning of this instruction! */
/* that is before the instruction is written */

/* length 4 or 8  */
void compile_gotoq(wb_t wb, op_t * op)
{
  uval length1, length2;
  uval opbyte;

  if (op->label->offset == UVAL_MAX) {/* this is to an uncompiled location */
    length1 = SVAL_MAX;
    length2 = -(op->offset + 8);   /* this plus the offset  */
  } else {
    if (op->r1 != ZERO) {
      length1 = op->label->offset - (op->offset + 4);
      length2 = op->label->offset - (op->offset + 8);
    } else {
      length1 = op->label->offset - (op->offset + 2);
      length2 = op->label->offset - (op->offset + 6);
    }
  }
  
  opbyte = 0x78;
  if (op->r1 != ZERO)  {
    wb(0x85);  /* test op->r1, op->r1 */
    wb(calc_mod_byte(3,op->r1,op->r1));
    opbyte = 0x74;
  }

  if (is_short_offset(length1)) {
    wb(opbyte);  /* jz/js address */
    wb(short_offset(length1));
  } else {
    addr_un repack;
    repack.value = length2;
    wb(0x0f);  /* jz/js address */
    wb(opbyte + 0x10);
    wb(repack.data[0]);
    wb(repack.data[1]);
    wb(repack.data[2]);
    wb(repack.data[3]);
  }
}

/* length 1 */
void compile_ret(wb_t wb, op_t *op)
{
  wb(0xC3);
}

/* 2 to 7 bytes */
void compile_load(wb_t wb, op_t * op)
{
  if (op->r1 == ZERO) return;  /* this is a nop */

  if (op->r2 == ZERO)   /* was iload */
    {
      addr_un repack;
      repack.value = long_offset(op->value);
      
      wb(0x8B);
      wb(calc_mod_byte(contents_of,disp32,op->r1));
      wb(repack.data[0]);
      wb(repack.data[1]);
      wb(repack.data[2]);
      wb(repack.data[3]);
    }
  else if ((op->value == 0) && (op->r2 != EBP))   /* was load [r2] */
    {
      wb(0x8B);  /* mov r1 <- [r2] */
      wb(calc_mod_byte(contents_of, op->r2, op->r1));
      if (op->r2 == ESP)
	wb(calc_scale_byte(times1,no_index,op->r2));
    }
      /* the general case, new code */
  else if ( is_short_offset(op->value))
    {
      wb(0x8b);
      wb(calc_mod_byte(contents_disp8_reg,op->r2, op->r1));
      if (op->r2 == ESP)  wb(calc_scale_byte(times1,no_index,op->r2));
      wb(short_offset(op->value));
    }
  else 
    {
      addr_un repack;
      repack.value = long_offset(op->value);
      
      wb(0x8b);
      wb(calc_mod_byte(contents_disp32_reg, op->r2, op->r1));
      if (op->r2 == ESP)  wb(calc_scale_byte(times1,no_index,op->r2));
      
      wb(repack.data[0]);
      wb(repack.data[1]);
      wb(repack.data[2]);
      wb(repack.data[3]);
    }
} 


/* length 2 to 7 */
void compile_store(wb_t wb, op_t * op)
{
  if (op->r2 == ZERO)  /* store zero in the location */
    {
      addr_un repack;
      repack.value = op->value;
      wb(0xC7);
      if ((op->value == 0) && (op->r1 != ZERO))
	wb(calc_mod_byte(contents_of,op->r2,0));
      else if (is_short_offset(op->value)) {
	wb(calc_mod_byte(contents_disp8_reg,op->r2,0));
	wb(short_offset(op->value));
      } else {
	wb(calc_mod_byte(contents_disp32_reg,op->r2,0));
	wb(repack.data[0]);
	wb(repack.data[1]);
	wb(repack.data[2]);
	wb(repack.data[3]);
      }
      wb(0);  /* ugh */
      wb(0);  /* ick */
      wb(0);  /* groan */
      wb(0);  /* gak  */
    }
  else if (op->r1 == ZERO)  /* was istore */
    {
      addr_un repack;
      repack.value = op->value;
      
      wb(0x89);
      wb(calc_mod_byte(contents_of,disp32,op->r2));
      wb(repack.data[0]);
      wb(repack.data[1]);
      wb(repack.data[2]);
      wb(repack.data[3]);
    }
 else if ((op->value == 0) && (op->r1 != EBP))
    {
      wb(0x89); /* mov [r1] <- r2 */
      wb(calc_mod_byte(contents_of, op->r1, op->r2));
      if (op->r1 == ESP) wb(calc_scale_byte(times1,no_index,op->r1));
    }
      /* the general case, new code */
  else if ( is_short_offset(op->value))
    {
      wb(0x89);
      wb(calc_mod_byte(contents_disp8_reg,op->r1, op->r2));
      if (op->r1 == ESP)  wb(calc_scale_byte(times1,no_index,op->r1));
      wb(short_offset(op->value));
    }
  else 
    {
      addr_un repack;
      repack.value = long_offset(op->value);
      
      wb(0x89);
      wb(calc_mod_byte(contents_disp32_reg, op->r1, op->r2));
      if (op->r1 == ESP)  wb(calc_scale_byte(times1,no_index,op->r1));
      
      wb(repack.data[0]);
      wb(repack.data[1]);
      wb(repack.data[2]);
      wb(repack.data[3]);
    }
}

/* length 5 */
void compile_set (wb_t wb, op_t *op)
{
  addr_un repack;
  repack.value = op->value;
  
  if (op->r1 == ZERO) return;  /* can't set zero */

  if (op->r2 == ZERO)
    {
      wb(0xB8 + op->r1);  /* mov r1,d32 */
      wb(repack.data[0]);
      wb(repack.data[1]);
      wb(repack.data[2]);
      wb(repack.data[3]);
    }
  else if (op->value == 0)
    {
      wb(0x89);   /* mov r1,r2 */
      wb(calc_mod_byte(value_of, op->r1, op->r2));
    }
  else 
    {
      wb(0x8D); /* lea r1, [r2 + value] */
      if (is_short_offset(op->value))  /* I might kill this case for relocation reasons */
	{
	  wb(calc_mod_byte(contents_disp8_reg, op->r2, op->r1));
	  wb(short_offset(op->value));
	}
      else
	{
	  wb(calc_mod_byte(contents_disp32_reg, op->r2, op->r1));
	  wb(repack.data[0]);
	  wb(repack.data[1]);
	  wb(repack.data[2]);
	  wb(repack.data[3]);
	}
    }
}

/* length 0,2 */
void compile_xchg(wb_t wb,op_t * op)
{
  reg_t r1,r2;
  if (op->r1 == op->r2) return;  /* thats a noop */
  
  if (op->r1 > op->r2) {      /* put the lowest numbered register in r1 */
    r1 = op->r2; r2 = op->r1;
  } else { 
    r1 = op->r1; r2 = op->r2;
  }
  if ((r1 == EAX) && (r2 != ZERO)) {
    wb(0x90+r2);  /* xchg EAX, r2 */
  } else if (r2 != ZERO) {
    wb(0x87);     /* xchg r1, r2 */
    wb(calc_mod_byte(value_of, r1, r2));
  } else {
    wb(0x31);     /* xor r1,r1 */
    wb(calc_mod_byte(value_of, r1, r1));
  }
}

/* length 1,2,3,5,6 or 8 */
void compile_add(wb_t wb, op_t * op)
{
  reg_t r1,r2;
  addr_un repack;
  repack.value = op->value;
  r1 = op->r1;
  r2 = op->r2;

  if ((r1 == ZERO) || ((r2 == ZERO) && (op->value == 0)))
    return; /* stupid noop */
    
  if ((r1 == r2) && (r2 == ESP) && (op->value != 0))
    {   /* lea can't handle this one */
      wb(0x01);               /* add ESP,ESP */
      wb(calc_mod_byte(value_of, r1, r2));
      r2 = ZERO;        /* later  add ESP,imm or lea ESP,[ESP + value] */
    }
  
  if (op->value == 0) {
      wb(0x01);
      wb(calc_mod_byte(value_of, r1, r2));
  } else if ((r2 == ZERO) && (op->value == 1)) {
      wb(0x40 + r1);  /* inc r1 */
  } else if ((r2 == ZERO) && (op->value == 0xFFFFFFFF)) {
      wb(0x48 + r1);  /* dec r1 */
  } else if (r2 == ZERO)  /* was addi */
    {
      if (r1 == EAX) {
	wb(0x05);
      } else {
	wb(0x81);
	wb(calc_mod_byte(value_of,op->r1,add));
      }
      wb(repack.data[0]);
      wb(repack.data[1]);
      wb(repack.data[2]);
      wb(repack.data[3]);
    }
  else if (is_short_offset(op->value))  
    /* lea takes 2 clocks but is smaller then 2 instructions */
    {
      reg_t reg1, reg2;
      reg2 = (r2 == ZERO)? no_index : ((r2 == ESP)? r1: r2);
      reg1 = (r2 == ESP)? r2 : r1;
      wb(0x8D);
      if (is_short_offset(op->value)) {
	wb(calc_mod_byte(contents_disp8_reg,scaled_index, r1));
	wb(calc_scale_byte(times1, reg2 , reg1));
	wb(short_offset(op->value));
      } else {
	wb(calc_mod_byte(contents_disp32_reg,scaled_index, r1));
	wb(calc_scale_byte(times1, reg2, reg1));
	wb(repack.data[0]);
	wb(repack.data[1]);
	wb(repack.data[2]);
	wb(repack.data[3]);
      }
    }
}



/* length 2 */
void compile_sub(wb_t wb, op_t *op)
{
  addr_un repack;
  repack.value = op->value;

  if ((op->r1 == ZERO) || ((op->r2 == ZERO) && (op->value == 0)))
    return; /* stupid noop */

  if (op->r2 != ZERO)
    {
      wb(0x29);  /* sub op->r1, op->r2 */
      wb(calc_mod_byte(value_of,op->r1,op->r2));
    }
  if (op->value == 0) return;

  if (op->value == 1) {
    wb(0x48 + op->r1);  /* dec op->r1 */
  } else if (op->value == 0xFFFFFFFF) {
    wb(0x40 + op->r1);  /* inc op->r1 */
  } else {
    if (op->r1 == EAX) 
      wb(0x2D);  /* sub EAX, [ ] */
    else {
      wb(0x81);  /* sub op->r1, [ ] */
      wb(calc_mod_byte(value_of,op->r1,sub));
    }
    wb(repack.data[0]);
    wb(repack.data[1]);
    wb(repack.data[2]);
    wb(repack.data[3]);
  }
}

/* length 1 */
void output_nop(wb_t wb, op_t *op)
{
  wb(0x90);  /* xchg eax,eax */ /* that's a nop */
}

/* length 0 */
void compile_nop(wb_t wb, op_t *op)
{
#if EXPLICIT_NOPS
  output_nop(wb,op);
#endif
}

/* the value parameter is the interrupt number, for the 80x86 */
/* for the mips it would be the value embeded in the instruction that is ignored */
/* mostly this value is unused or the same throughout but, */
/* it is a possibly useful hook so I provide it */

void compile_syscall(wb_t wb, op_t *op)
{
  if (op->value > 255) {
    fprintf(stderr,"attempt to compile and invalid system call number %u\n!!",op->value);
    exit(EXIT_FAILURE);
  } else if (op->value == 3 ) {
    wb (0xCC);
  } else {
    wb(0xCD);
    wb(op->value &0xFF);
  }
}


/*---------------- Special cases ---------------- */
void compile_pushl(wb_t wb, op_t *op)
{
  wb(0x50 + op->next->r2);  /* pushl r2 */
}

op_t *is_pushl(op_t *op)
{
  /* pushl is 
    sub ESP,4
    store [ESP],reg
  */
  int value;
  if (op->next == 0) {
    value = 0;
  } else if (op->next->opcode == STORE)
    {
      if (op->opcode == SUB)
	value = op->value == 4;
      else if (op->opcode == ADD)
	value = op->value == 0xFFFFFFFC;
      else
	value = 0;
      value = value && (op->r1 == ESP) && (op->r2 == ZERO);
      value = value && (op->next->r1 == ESP) && (op->next->value == 0);
    } 
  else 
    value = 0;
  
  return value? op->next->next : 0;
}

void compile_popl(wb_t wb, op_t *op)
{
  wb(0x58 + op->r1); /* popl r1 */
}

op_t *is_popl(op_t *op)
{
  /* popl is 
     load reg,[ESP]
     add ESP,4
   */
  int value;
  if (op->next == 0) {
    value = 0;
  } else if (op->opcode == LOAD)
    {
      if (op->next->opcode == ADD)
	value = op->next->value == 4;
      else if (op->next->opcode == SUB)
	value = op->next->value == 0xFFFFFFFC;
      else
	value = 0;
      value = value && (op->r2 == ESP) && (op->value == 0);
      value = value && (op->next->r1 == ESP) && (op->next->r2 == ZERO);
    } 
  else
    value = 0;

  return value? op->next->next:0;
}


