#include "error.h"
#include "spaghetti-stack.h"
#include "stack.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

/* --------- error handling ------------ */
void do_stack_fatal_error(char *format, ... )
{
  va_list ap;
  va_start(ap,format);
  vfprintf(stderr,format,ap);
  va_end(ap);
  exit(EXIT_FAILURE);
}

#ifndef C
#define C(X) 
#endif

/* the routines contained herein should be definitions of
   correctness, not implementations of efficiency */

/*------------- pointer comparisons -------------*/
/* this one is ridicolous */
static cell equal(cell *first, cell *second, stack_data * stack) 
{
  return first == second;
}

static cell above(cell *first, cell *second, stack_data * stack)
{
  cell flag;
  if (stack->direction == up)
    {
      flag = first > second;
    }
  else /* stack->direction == down */
    {
      flag = first < second;
    }
  return flag;
}

static cell below(cell *first, cell *second, stack_data * stack)
{
  cell flag;
  if (stack->direction == up)
    {
      flag = first < second;
    }
  else /* stack->direction == down */
    {
      flag =  first > second;
    }
  return flag;
}

static scell difference(cell *first, cell *second, stack_data *stack)
{
  scell diff;
  if (stack->direction == up)
    {
      diff = first - second;
    }
  else /* if (stack->direction == down) */
    {
      diff = second - first;
    }
  return diff;
}

/*------------- cheap stack routines ------------*/
/* with no checking of any kind */
void cheap_push(cell val, stack_data *stack)
{
  (stack->place_pushed) = (stack->place_pushed) + (stack->direction);
  *(stack->place_pushed) = val;
  stack->place_to_pop = stack->place_pushed;
}

cell cheap_pop(stack_data *stack)
{
  cell temp;
  temp = *(stack->place_to_pop);
  (stack->place_to_pop) = (stack->place_to_pop) - (stack->direction);
  return temp;
}

C( 5 4 3 2 1 0 )  /* items selected by various offsets */
cell cheap_peek(cell offset, stack_data *stack)  /* offset is positive */
{
  cell *place;
  place = (stack->place_to_pop) - (offset * (stack->direction));
  return *place;
}

C( 5 4 3 2 1 0 ) /* items changes by various offsets */
void cheap_poke(cell val, cell offset, stack_data *stack) /* offset is positive */
{
  cell *place;
  place = (stack->place_to_pop) - (offset * (stack->direction));
  *place = val;
}


/*------------- cheap-safe routines ------------*/
/* these check stack boundaries top & bottom */

void cheap_safe_push(cell val, stack_data *stack)
{
  if (equal(stack->place_pushed,stack->top,stack))
    {
      do_stack_fatal_error("%s stack overflowing aborting!!\n",stack->name);
    }
  else
    cheap_push(val,stack);
}

cell cheap_safe_pop(stack_data *stack)
{
  cell * place_to_push = (stack->place_to_pop) + (stack->direction);
  if (equal(place_to_push, stack->bottom, stack))  /* i.e. if stack is empty I can't pop */
    {
      do_stack_fatal_error("%s stack underflowing, aborting!!\n",stack->name);
    }
  return cheap_pop(stack);
}

cell cheap_safe_peek(cell offset, stack_data *stack)
{
  cell *place;
  place = (stack->place_to_pop) - ((offset) * (stack->direction));
  
  if ( below( place, stack->bottom, stack ) )
    {
      do_stack_fatal_error("attempt to peek under bottom of %s stack, aborting!!\n",stack->name);
    }
  else
    {
      return cheap_peek(offset,stack);
    }
}

void cheap_safe_poke(cell val, cell offset, stack_data *stack)
{
  cell *place;
  place = (stack->place_to_pop) - ((offset) * (stack->direction));
  
  if ( below( place, stack->bottom, stack ) )
     {
      do_stack_fatal_error("attempt to poke under bottom of %s stack, aborting!!\n",stack->name);
    }
  else
    {
      cheap_poke(val, offset, stack);
    }
}

/*---------------- frame-routines --------------*/

static cell *get_frame_top(cell *frame, stack_data *stack)
{
  cell *top_ptr;
  top_ptr = frame - (FRAME_TOP_OFF*(stack->direction));
  return (cell *)(*top_ptr);
}

void frame_top_set(cell *val,cell *frame, stack_data *stack)
{
  cell *top_ptr;
  top_ptr = frame - (FRAME_TOP_OFF*(stack->direction));
  *top_ptr = (cell)(val);
}

cell *frame_previous(cell *frame, stack_data *stack)
{
  cell * prev_ptr;
  prev_ptr = frame - (FRAME_PREV_OFF*(stack->direction));
  return (cell *)(*prev_ptr);
}

void frame_previous_set(cell *val, cell *frame, stack_data *stack)
{
  cell * prev_ptr;
  prev_ptr = frame - (FRAME_PREV_OFF*(stack->direction));
  *prev_ptr = (cell)(val);
}

cell is_top_frame(cell *frame, stack_data *stack)
{
  cell *top;
  cell flag;
  top = get_frame_top(frame,stack);

  flag = equal(top,stack->top,stack);
  return flag;
}

cell *frame_top(cell *frame, stack_data *stack)
{
  cell *top;
  if (is_top_frame(frame,stack))
    top = stack->place_pushed;
  else
    top = get_frame_top(frame,stack);
  return top;
}

cell is_bottom_frame(cell *frame, stack_data *stack)
{
  cell * frame_bottom;
  frame_bottom = frame - (FRAME_DEPTH * (stack->direction));
  return equal(stack->bottom,frame_bottom,stack);
}

void push_frame(stack_data *stack)
{
  cell * previous;
  unsigned i;
  previous = stack->frame;
  if ((previous != 0) && is_top_frame(previous, stack))
    {
      frame_top_set(stack->place_pushed, previous, stack);
    }
  
  for(i = 0; i < FRAME_DEPTH; i++)
    cheap_safe_push(0,stack);  /* create space for the stack frame */
  
  stack->frame = (stack->place_pushed) + (stack->direction);
  frame_top_set(stack->top, stack->frame, stack);
  frame_previous_set(previous, stack->frame, stack);
}

void pop_frame(stack_data *stack)
{
  cell * frame;
  cell * previous;
  cell * place_to_push;
  cell *new_top;
 
  frame = stack->frame;
  
  if (is_bottom_frame(frame,stack))
    {
      do_stack_fatal_error("attempt to pop the first %s stack frame, aborting!!\n",stack->name);
    }

  previous = frame_previous(frame,stack);
  new_top = frame_top(previous,stack);
  stack->frame = previous;
  stack->place_to_pop = new_top;

  place_to_push = (stack->place_pushed) + (stack->direction);

/* only remove an empty stack frame */
/* stack->place_to_pop is bounded between,
   stack->place_pushed and stack->frame */
  
  if (is_top_frame(frame,stack) && equal(place_to_push, frame, stack))
    stack->place_pushed -= FRAME_DEPTH*(stack->direction);  /* remove the frame */

  if (equal(new_top, stack->place_pushed, stack))        /* the revealed frame is on top */
    frame_top_set(stack->top,previous,stack);
}

cell frame_depth(cell *frame, stack_data *stack)
{
  cell *top;
  cell depth;

  top = frame_top(frame,stack);
  depth = difference(stack->place_pushed, frame, stack) +1;
  return depth;
}


/*------------- frame-safe routines ------------*/

void frame_safe_sync(stack_data *stack)
{
  if (is_top_frame(stack->frame,stack))
    {
      stack->place_pushed = stack->place_to_pop;
    }
}

void frame_safe_push(cell val, stack_data *stack)
{
  if (!is_top_frame(stack->frame,stack))
    push_frame(stack);
  cheap_safe_push(val,stack);
}

cell frame_safe_pop(stack_data *stack)
{
  cell val;
  cell * place_to_push;

  place_to_push = (stack->place_to_pop) + (stack->direction);
  while(equal(place_to_push, stack->frame, stack))  /* i.e. if stack frame is empty I can't pop */
   {
      pop_frame(stack);  /* aborts if I try to pop the first stack frame */
      place_to_push = (stack->place_to_pop) + (stack->direction);
    } 

  val = cheap_safe_pop(stack);
  frame_safe_sync(stack);
  return val;
}

cell frame_safe_peek(cell offset, stack_data *stack)
{
  cell *local_frame;
  cell depth;
  
  local_frame = stack->frame;
  depth = frame_depth(local_frame,stack);

  while((offset > depth) && !is_bottom_frame(local_frame,stack))
    {
      offset -= depth;
      local_frame = frame_previous(local_frame,stack);
      depth = frame_depth(local_frame,stack);
    }
  if (offset > depth)
    {
      do_stack_fatal_error("attempt to peek under bottom of %s stack, aborting!!\n",stack->name);
    }
  else
    {  /* value is guarenteed to be in this frame */
      cell *place;
      place = frame_top(local_frame,stack) - (offset * (stack->direction));
      return *place;
    }
}

/* this may need more work to allow it to be equivalent to 
   pushes and pops,  That is if I poke in to a frame thats
   not on the top of the stack I need to generate new frames 
   etc etc,
   For now beware that poke is destructive 
 */
void frame_safe_poke(cell val, cell offset, stack_data *stack)
{
  cell *local_frame;
  cell depth;
  
  local_frame = stack->frame;
  depth = frame_depth(local_frame,stack);

  while((offset > depth) && !is_bottom_frame(local_frame,stack))
    {
      offset -= depth;
      local_frame = frame_previous(local_frame,stack);
      depth = frame_depth(local_frame,stack);
    }
  if (offset > depth)
    {
      do_stack_fatal_error("attempt to poke under bottom of %s stack, aborting!!\n",stack->name);
    }
  else
    {  /* value is guarenteed to be in this frame */
      cell *place;
      place = frame_top(local_frame,stack) - (offset * (stack->direction));
      *place = val;
    }
}

cell frame_safe_depth(stack_data *stack)
{
  cell *local_frame;
  cell depth;
  
  local_frame = stack->frame;
  depth = frame_depth(local_frame,stack);

  while(!is_bottom_frame(local_frame,stack))
    {
      local_frame = frame_previous(local_frame,stack);
      depth += frame_depth(local_frame,stack);
    }
  return depth;
}

/*----------- smile ---------------*/


/*---------- user-interface routines ---------------*/

void push(cell val, stack_data *stack)
{
  frame_safe_push(val,stack);
}

cell pop(stack_data *stack)
{
  return frame_safe_pop(stack);
}

cell peek(cell place, stack_data *stack)
{
  return frame_safe_peek(place, stack);
}

void poke(cell place, cell value, stack_data *stack)
{
  frame_safe_poke(value, place, stack);
}

cell depth(stack_data *stack)
{
  return frame_safe_depth(stack);
}

void stack_init(char *name, cell *data, cell length, sdir_t direction, stack_data *stack)
{
  stack->direction = direction;
  stack->name = name;
  if (direction == up)
    {
      stack->bottom = data;
      stack->top = &data[length -1];
    }
  else  /* if (direction == down) */
    {
      stack->top = data;
      stack->bottom = &data[length -1];
    }
  
  stack->place_pushed = stack->bottom - (stack->direction);
  stack->place_to_pop = stack->bottom - (stack->direction);
  stack->frame = 0;
  push_frame(stack);
}
