#ifndef SPAGHETTI_STACK_H
#define SPAGHETTI_STACK_H

typedef unsigned long int cell;
typedef signed   long int scell;
#define PTOC(X) ((cell)(X))
#define CTOP(X) ((cell *)(X))

/* note for now at least stacks grow toward numerically smaller addresses (down) */
/* just like the return stack on the 80x86 (so it can be a standard stack) */

typedef enum stack_dir { down = -1, up = 1 } sdir_t;
typedef struct stack_data {
  sdir_t direction;      /* the direction the stack grows (pushes) */
  char * name;           /* for debugging, & error messages */
  cell * top;            /* the last storage address for pushes */
  cell * bottom;         /* the last storage address for pops */
  cell * frame;          /* address of a stack discontinuity for pops */
  cell * place_pushed;   /* with this you find where to push next */
  cell * place_to_pop;   /* with this you find where to pop next */
}stack_data;

/* The stack is an implmentation of a tree stack.
   Tree stacks should be cheaper to use for implementing
   continuations, then copying the whole stack.
   There are also some special GC algorithms for tree stacks, are
   cheaper then GC'ing the data space. 
   I hope to test tree stacks to see if they can deliver so I'll
   implement them first. 
   If they aren't effeciant enough we can try other kinds.

   An important question is how to implement them into compiled code.

   A guard page at the top & bottom can help detect stack overflow and
   underflow, so those 2 conditions I'll not worry about in compiled
   code. 

   Registers though are quite few (esp. on the 80x86) and with several
   stacks keeping two stack pointers in memory could be prohibitive.
   It is my goal to be able to use only one pointer for the common
   case.  Keeping everything else in memory.

   How to compile correct, efficient code is a major question for this
   approach.  But also compiling efficient code is a major question in
   general for me at least.  If you can't compile this efficiently
   this scheme get scrapped.  But that's a few revisions away yet.
*/

#define FRAME_TOP_OFF 1
#define FRAME_PREV_OFF 2
#define FRAME_DEPTH 2

/* structure of a stack frame */
/*
 *  frame_pointer[-FRAME_TOP_OFF ] == top of the frame
 *  frame_pointer[-FRAME_PREV_OFF] == previous frame pointer
 */

/* NOTE NOTE NOTE  -- IDEA IDEA IDEA
 * if
 * frame_pointer[-3] == frame_limit (or something like that)
 * I can expand the basic concept to have no contiguous stack blocks
 * The basic price is already paid anyway because of the ability to
 * grab continuations.  This should be waid against ease of garbage
 * collecting particular stacks, speed, and cache access.
 * I think if stack blocks were allocated upon a general heap, with a
 * grain, and independently garbage collected (as well as dependently)
 * it would work well.  The important question to ask is how well does
 * this work for a light weight garbage collector for the stacks, and
 * especially how well does this work for the memory cache.
 */
/*------------- cheap stack routines ----------- */
void cheap_push(cell val, stack_data *stack);
cell cheap_pop(stack_data *stack);
cell cheap_peek(cell offset, stack_data *stack);
void cheap_poke(cell val, cell offset, stack_data *stack);

/*-------------- cheap-safe routines -------------*/
void cheap_safe_push(cell val, stack_data *stack);
cell cheap_safe_pop(stack_data *stack);
cell cheap_safe_peek(cell offset, stack_data *stack);
void cheap_safe_poke(cell val, cell offset, stack_data *stack);

/*---------------- frame-routines --------------*/
cell *frame_top(cell *frame, stack_data *stack);
void frame_top_set(cell *val,cell *frame, stack_data *stack);
cell *frame_previous(cell *frame, stack_data *stack);
void frame_previous_set(cell *val, cell *frame, stack_data *stack);
cell is_top_frame(cell *frame, stack_data *stack);
cell is_bottom_frame(cell *frame, stack_data *stack);
void push_frame(stack_data *stack);
void pop_frame(stack_data *stack);
cell frame_depth(cell *frame, stack_data *stack);

/*------------- frame-safe routines ------------*/
void frame_safe_push(cell val, stack_data *stack);
cell frame_safe_pop(stack_data *stack);
cell frame_safe_peek(cell offset, stack_data *stack);
void frame_safe_poke(cell val, cell offset, stack_data *stack);
cell frame_safe_depth(stack_data *stack);

/*------------- user interface ----------------- */
void push(cell val, stack_data * stack);
cell pop(stack_data *stack);
cell peek(cell place, stack_data * stack);
void poke(cell place, cell value, stack_data *stack);
cell depth(stack_data *stack);
void stack_init(char *name, cell *data, cell length, sdir_t direction, stack_data *stack);
#endif SPAGHETTI_STACK_H


